from dataclasses import dataclass, field
import inspect
from enum import Enum, IntEnum
from inspect import Signature, BoundArguments, Parameter
from typing import Any, Callable, ParamSpec, TypeVar, Sized, cast
from src.verliberr import ErrKind, ErrMsg, VerLibErr
from src.jsonrpc import (
    Error,
    ErrorCode,
    ErrorMsg,
    JSONValues,
    Request,
    Response,
    OkRes,
    ErrRes,
)
from utils.result import Err, Ok, Result

P = ParamSpec("P")
T = TypeVar("T", bound=JSONValues)


VerProc = Callable[P, T]
VerProcParams = JSONValues


@dataclass
class VerProcedure:
    name: str
    _fn: Callable[..., JSONValues]
    _signature: Signature

    def call(
        self, params: list[JSONValues] | dict[str, JSONValues] | None
    ) -> Result[JSONValues, VerLibErr]:

        pos_params_len = len(
            tuple(
                filter(
                    lambda p: p.kind == Parameter.POSITIONAL_OR_KEYWORD,
                    self._signature.parameters.values(),
                )
            )
        )

        if params is None and pos_params_len > 0:
            return Err(VerLibErr(ErrKind.INVALID_PARAMS, ErrMsg.INVALID_PARAMS))

        if params is not None and len(params) < pos_params_len:
            return Err(VerLibErr(ErrKind.INVALID_PARAMS, ErrMsg.INVALID_PARAMS))

        # TODO: Make sure function can be called with 'null' arg
        if params is None or len(params) == 0:
            return Ok(self._fn())

        args, kwargs = ([], {})
        match params:
            case list(pos_params):
                args = pos_params
            case dict(dict_params):
                kwargs = dict_params
        # TODO: Add type checking for parameters
        try:
            ba = self._signature.bind(*args, **kwargs)
        except TypeError:
            return Err(VerLibErr(ErrKind.INVALID_PARAMS, ErrMsg.INVALID_PARAMS))

        # TODO: Decide on the best way to deal with exceptions
        return Ok(self._fn(*ba.args, **ba.kwargs))


@dataclass
class VerModule:
    name: str
    _procedures: dict[str, VerProcedure]

    def __init__(self, name: str):
        self.name = name
        self._procedures = {}

    def _register_proc(self, proc: VerProcedure):
        self._procedures[proc.name] = proc

    def verproc(
        self, fn: VerProc[P, T] | None = None, *, name: str = ""
    ) -> Callable[[VerProc[P, T]], VerProc[P, T]] | VerProc[P, T]:
        def verproc_decorator(procedure: VerProc[P, T]) -> VerProc[P, T]:
            proc_name = name if name != "" else procedure.__name__
            if proc_name in self._procedures:
                raise TypeError(
                    f"A procedure with the name '{proc_name}' has already been registered to the module '{self.name}'"
                )

            self._register_proc(
                VerProcedure(proc_name, procedure, inspect.signature(procedure))
            )
            return procedure

        if fn is None:
            return verproc_decorator
        else:
            return verproc_decorator(fn)

    def _contains_proc(self, name: str) -> bool:
        return name in self._procedures

    def _call_procedure(
        self,
        proc_name: str,
        params: list[JSONValues] | dict[str, JSONValues] | None,
    ) -> Result[JSONValues, VerLibErr]:
        return self._procedures[proc_name].call(params)


@dataclass
class VerLib:
    name: str
    _modules: dict[str, VerModule]

    def __init__(self, name: str):
        self.name = name
        self._default_module: VerModule = VerModule("_default_")
        self._modules = {}

    def declare_module(self, module: VerModule):
        mod_name = module.name
        if mod_name in self._modules:
            raise TypeError(
                f"A module with the name '{mod_name}' has already been declared"
            )
        self._modules[mod_name] = module

    def verproc(
        self, fn: VerProc[P, T] | None = None, *, name: str = ""
    ) -> Callable[[VerProc[P, T]], VerProc[P, T]] | VerProc[P, T]:

        return self._default_module.verproc(fn, name=name)

    def _resolve_proc(self, name: str) -> tuple[VerModule | None, str]:
        components = name.split(".")
        if len(components) == 1:
            return (self._default_module, components[0])
        elif len(components) == 2:
            return (
                self._modules.get(components[0]),
                components[1],
            )
        else:
            return (None, "")

    def execute_rpc(self, req: Request) -> Response[JSONValues, None]:
        # Check if module and method both exist
        module, proc_name = self._resolve_proc(req.method)
        if not module or not module._contains_proc(proc_name):
            return ErrRes(
                req.id,
                Error(
                    ErrorCode.METHOD_NOT_FOUND,
                    ErrorMsg.METHOD_NOT_FOUND.format(req.method),
                    None,
                ),
            )

        result: Result[JSONValues, VerLibErr] = module._call_procedure(
            proc_name, req.params
        )

        ignore_result: bool = req.is_notification
        if result.is_ok():
            return OkRes(req.id, result.unwrap() if not ignore_result else None)

        return result.unwrap_err().into_json_rpc_err(req.id)



from __future__ import annotations
import dataclasses
import abc
from dataclasses import dataclass
from typing import Literal, Any, TypeVar, Generic, cast
from enum import Enum, IntEnum
import json
from schema import Schema, And, Or, Optional
from typing_extensions import Self
from utils.result import Result, Ok, Err


JSONValues = (
    int | str | float | list["JSONValues"] | dict[str, "JSONValues"] | None
)


JSONRPCId = int | str | None
JSONRPCParams = dict[str, JSONValues] | list[JSONValues] | None

V = TypeVar("V")

E = TypeVar("E")

values_schema = Schema(Or(int, str, float, list, dict, lambda x: x is None))

request_schema = Schema(
    {
        "jsonrpc": "2.0",
        Optional("id"): Or(int, str, lambda x: x is None),
        "method": str,
        Optional("params"): Or(
            {str: values_schema}, And([values_schema], lambda l: len(l) > 0)
        ),
    }
)


@dataclass(kw_only=True)
class Request:
    jsonrpc: Literal["2.0"] = "2.0"
    method: str
    is_notification: bool
    id: int | str | None = None
    params: JSONRPCParams = None

    def __init__(
        self, *, method: str, id: JSONRPCId = None, params: JSONRPCParams = None
    ):
        self.method = method
        self.id = id
        self.params = params
        self.is_notification = id is None

    @classmethod
    def from_dict(cls, req_dict: dict[str, Any]) -> Self:
        request_schema.validate(req_dict)
        req_dict.pop("jsonrpc", None)
        return cls(**req_dict)


class ErrorCode(IntEnum):
    PARSE_ERROR = -32700
    INVALID_REQUEST = -32600
    METHOD_NOT_FOUND = -32601
    INVALID_PARAMS = -32602
    INTERNAL_ERROR = -32603


class ErrorMsg(Enum):
    METHOD_NOT_FOUND = "The procedure '{0}' was not found on the server."
    INVALID_PARAMS = "Invalid method parameter(s)."

    def format(self, *args: Any, **kwargs: Any) -> str:
        return self.value.format(*args, **kwargs)


@dataclass
class Error(Generic[E]):
    code: ErrorCode | int
    message: str
    data: E


@dataclass
class OkRes(Generic[V]):
    id: JSONRPCId
    result: V
    jsonrpc: Literal["2.0"] = "2.0"

    def __init__(self, id: JSONRPCId, result: V):
        self.id = id
        self.result = result

    def err_data(self) -> None:
        return None

    def result_data(self) -> V:
        return self.result

    def is_err(self) -> bool:
        return False

    def is_success(self) -> bool:
        return True

    def to_dict(self) -> dict[str, JSONValues]:
        return dataclasses.asdict(self)


@dataclass
class ErrRes(Generic[E]):
    id: JSONRPCId
    error: Error[E]
    jsonrpc: Literal["2.0"] = "2.0"

    def __init__(self, id: JSONRPCId, error: Error[E]):
        self.id = id
        if error.code in (ErrorCode.PARSE_ERROR, ErrorCode.INVALID_REQUEST):
            self.id = None
        self.error = error

    def err_data(self) -> Error[E]:
        return self.error

    def result_data(self) -> None:
        return None

    def is_err(self) -> bool:
        return True

    def is_success(self) -> bool:
        return False

    def to_dict(self) -> dict[str, JSONValues]:
        return dataclasses.asdict(self)


Response = OkRes[V] | ErrRes[E]


def into_rpc_request(
    req: str | dict[str, JSONValues]
) -> Result[Request, Error[None]]:

    try:
        req_dict: dict[str, Any] = (
            req if isinstance(req, dict) else json.loads(req)
        )
    except TypeError:
        return Err(Error(ErrorCode.PARSE_ERROR, "Parse error", None))

    is_valid_req = request_schema.is_valid(req_dict)

    if not is_valid_req:
        return Err(Error(ErrorCode.INVALID_REQUEST, "Invalid Request", None))

    return Ok(Request.from_dict(req_dict))



from src.verlib import VerLib
from src.jsonrpc import Request, ErrRes, OkRes
import src.jsonrpc as jsonrpc
import typing
from flask import Flask, request
import flask


class FlaskVerLib:
    def __init__(self, verlib: VerLib, api_url="/verlib"):
        self._verlib: VerLib = verlib
        self.api_url = api_url

    def init_app(self, app: Flask):
        self._dispatch_rpc_call = app.post(self.api_url)(
            self._dispatch_rpc_call
        )

    def _dispatch_rpc_call(self) -> flask.Response:

        req_json: dict = request.get_json()
        req_id = req_json.get("id")
        rpc_req = jsonrpc.into_rpc_request(request.get_json())

        if rpc_req.is_err():
            return flask.jsonify(ErrRes(req_id, rpc_req.unwrap_err()))

        rpc_call = rpc_req.unwrap()
        result = self._verlib.execute_rpc(rpc_call)

        if result.is_err():
            return flask.jsonify(
                ErrRes(rpc_call.id, typing.cast(ErrRes, result).err_data())
            )
        else:
            return flask.jsonify(OkRes(rpc_call.id, result.result_data()))



from enum import IntEnum, Enum
import traceback as tb
from io import StringIO
from dataclasses import dataclass
from src.jsonrpc import (
    ErrRes,
    JSONRPCId,
    ErrorCode as JErrorCode,
    ErrorMsg as JErrorMsg,
    Error as JError,
)
import src.jsonrpc as jsonrpc


class ErrKind(IntEnum):
    INVALID_PARAMS = 0
    PROCEDURE_RAISED_EXCEPTION = -32500


class ErrMsg(Enum):
    INVALID_PARAMS = "Invalid procedure parameter(s)."
    PROCEDURE_RAISED_EXCEPTION = (
        "An error occurred during the execution of the procedure."
    )

    def __str__(self) -> str:
        return self.value


@dataclass
class VerLibErr:
    err_kind: ErrKind
    msg: ErrMsg
    exc: Exception | None = None

    def into_json_rpc_err(
        self,
        req_id: JSONRPCId,
    ) -> ErrRes:
        match self.err_kind:
            case ErrKind.INVALID_PARAMS:
                return ErrRes(
                    req_id,
                    JError(
                        JErrorCode.INVALID_PARAMS,
                        JErrorMsg.INVALID_PARAMS.value,
                        None,
                    ),
                )
            case ErrKind.PROCEDURE_RAISED_EXCEPTION:
                # TODO: Design error treatment flow
                assert isinstance(self.exc, Exception)
                raise self.exc
            case _:
                return ErrRes(req_id, JError(-1, "", None))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['src', 'src.integrations']

package_data = \
{'': ['*']}

install_requires = \
['schema>=0.7.5,<0.8.0']

extras_require = \
{'django': ['django>=4.1.7,<5.0.0'], 'flask': ['flask>=2.2.3,<3.0.0']}

setup_kwargs = {
    'name': 'verlib-py',
    'version': '0.1.0',
    'description': 'library for building RPC APIs',
    'long_description': '',
    'author': 'stackswithans',
    'author_email': 'stexor12@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
